#include <string.h>
#include "hardwareprofile.h"
#include "asynctasks.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/timers.h"
#include "system.h"
#include "driver/uart.h"
#include "listx.h"
#include "pcb.h"
#include "utility.h"
#include "state.h"
#include "configuration.h"

#define TAG "Async tasks"
#define INTERVAL_CHECK_MS   20

#define DEPRESSED(self)     if (self->todie) kill_self(self);
//#define DEPRESSED(self)     //if (self->todie) vTaskDelay(1000/portTICK_PERIOD_MS);

myLIST_HEAD(process_list);
SemaphoreHandle_t async_process_sem;

void init_async_process_management() {
    async_process_sem = xSemaphoreCreateMutex(); 
    init_process_list();
}

void kill_self(pcb_t *self) {
    timer_args_t *params = self->args;
    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    out_procq(&process_list, self);
    free_pcb(self);
    free(params);
    xSemaphoreGive(async_process_sem);
    vTaskDelete(NULL);
}

void terminate_timer(pcb_t *self) {
    timer_args_t *params = self->args;
    uint8_t map;

    //clear_activity_bitmap_output(params->bitmap);
    send_output_state(STATE_CLEAR_ACTIVITY_OUTPUTS, params->bitmap, self->todie);
    map = params->dac >= 0 ? 1 << params-> dac : 0;
    //clear_activity_bitmap_dac(map);
    send_clear_activity_dac(map, self->todie);

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    ESP_LOGI(TAG, "Finished scaled dac task after %d milliseconds", params->timer*100);
    //kill_self(self);
    out_procq(&process_list, self);
    free_pcb(self);
    free(params);
    xSemaphoreGive(async_process_sem);
    vTaskDelete(NULL);
}


void _delete_task(pcb_t *todel) {

    ESP_LOGI(TAG, "deleteting process %d (%d) with alarm %i",(uint32_t) todel->task, (uint32_t)xTaskGetCurrentTaskHandle(), todel->alarm);
    ESP_LOGI(TAG, "about to delete");
    vTaskDelete(todel->task);
    ESP_LOGI(TAG, "deleted");
    free(todel->args);
    out_procq(&process_list, todel);
    free_pcb(todel);
    ESP_LOGI(TAG, "freed memory");
}

void __delete_task(pcb_t *todel) {

    // Imposta la cancellazione del processo
    todel->todie = 1;
    ESP_LOGI(TAG, "soft deleteting process %d with alarm %i",(uint32_t) todel, todel->alarm);
}


void delete_tasks_by_output(uint8_t *bitmap) {
    pcb_t *aux;
    int i, index = 0;
    timer_args_t *args;
    pcb_t *todelete[MAX_PROCESS];
    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    list_for_each_entry(aux, &process_list, p_next) {
        for (i = 0; i < TOT_OUTPUTS; i++) {
            args = (timer_args_t*) aux->args;
            if ((bitmap[i] & args->bitmap[i]) != 0) {
                ESP_LOGI(TAG, "scheduling task for deletion: bitmap");
                hexprint(args->bitmap, TOT_OUTPUTS, TAG);
                todelete[index++] = aux;
                break;
            }
        }
    }
    for (i = 0; i < index; i++)
        _delete_task(todelete[i]);
    xSemaphoreGive(async_process_sem);
    if (index > 0)
       vTaskDelay(1);
}

void delete_tasks_by_alarm(uint8_t *bitmap, uint16_t *ma420) {
    pcb_t  *aux;
    pcb_t *todelete[MAX_PROCESS];
    int index, i, num, bit;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    index = 0;
    list_for_each_entry(aux, &process_list, p_next) {
        if(aux->alarm >= 0) {
            num = aux->alarm/8;
            bit = aux->alarm%8;
            if (num < TOT_INPUTS && (bitmap[num] & 1 << bit) != 0) {
                ESP_LOGI(TAG, "scheduling process with alarm %i for deletion", aux->alarm);
                todelete[index++] = aux;
                continue;
            }
        }
    }
    for (i = 0; i < index; i++)
        _delete_task(todelete[i]);
    xSemaphoreGive(async_process_sem);
    if (index > 0)
        vTaskDelay(1);
}

void task_simple_simple_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    
    ESP_LOGI(TAG, "Started simple simple timer task for %i milliseconds", params->timer*100);
    send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);
    vTaskDelay(params->timer*100/portTICK_PERIOD_MS);

    terminate_timer(self);
}

void task_simple_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    uint8_t buffer[64];
    uint64_t finished;
    TickType_t delay;
    char firstreached = 0;
    
    ESP_LOGI(TAG, "Started timer task %d for %i milliseconds, limits %i and %i", (uint32_t)self, params->timer*100, params->first420limit, params->limit420ma);
    if (params->first420limit == 0)
        firstreached = 1;

    finished = get_millis() + params->timer*100;

    while(params->timer == 0 || get_millis() < finished) {
        delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis())/portTICK_PERIOD_MS;
        xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS, pdTRUE, pdFALSE, delay);
        DEPRESSED(self);
        ESP_LOGI(TAG, "value found: %i (%i) (%i %i)", 0, buffer[params->num420ma], params->first420limit, params->limit420ma);
        if (firstreached) { 
            ESP_LOGI(TAG, "Second limit reached at %llu", esp_timer_get_time());
            break;
        }
        else if (!firstreached && params->f_dac) {
            ESP_LOGI(TAG, "First limit reached!");
            firstreached = 1;
        }
    }

    terminate_timer(self);
}
void launch_simple_simple_timer_task(uint16_t timer, int alarm, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_simple_timer, "simple simple timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process (alarm %i)", process->alarm);
    xSemaphoreGive(async_process_sem);
}

void launch_simple_timer_task(uint16_t timer, int alarm, uint8_t limit420ma, int8_t num420ma, int direction, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->dac = -1;
    args->f_dac = 0;
    args->first420limit = 0;
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_timer, "simple timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process (alarm %i)", process->alarm);
    xSemaphoreGive(async_process_sem);
}

void launch_timer_task(uint16_t timer, int alarm, uint8_t limit420ma, int8_t num420ma, int direction, uint8_t numdac, 
    uint8_t dacstart, uint8_t dacfinish, uint8_t first420limit, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->dac = numdac;
    args->f_dac = 1;
    args->dacstart = dacstart;
    args->dacfinish = dacfinish;
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_timer, "timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process with ma420 limit %i (%i)", args->limit420ma, limit420ma);
    xSemaphoreGive(async_process_sem);
}

void get_output_timers(int num, uint16_t *timers) {
    int i;
    pcb_t *aux;
    timer_args_t *args;
    unsigned long time = get_millis();

    if (num == 155)
        ESP_LOGI(TAG, "using num");

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    for (i = 0; i < 8; i++) {
        timers[i] = 0;
        list_for_each_entry(aux, &process_list, p_next) {
            args = (timer_args_t*) aux->args; 
            if (args->bitmap[num] & (1<<i) && args->timer != 0) {
                if (args->timer == 0)
                    timers[i] = 0xFFFF;
                else
                    timers[i] = args->timer - (time-aux->timestamp)/100;
                break;
            }
        }
    }
    xSemaphoreGive(async_process_sem);
}

void task_timer_after_input(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    uint64_t finished, delay;

    char f_reached = 0;

    ESP_LOGI(TAG, "Started timer task (subordinated to input) for %i milliseconds", params->timer*100);
    finished = get_millis() + params->timer*100;

    // Un tempo nullo viene interpretato come operazione senza fine (assumendo di essere interrotto da fuori)
    while (params->timer == 0 || get_millis() < finished) {
        delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis())/portTICK_PERIOD_MS;
        xEventGroupWaitBits(params->eventgroup, INPUT_EVENT_BITS, pdTRUE, pdFALSE, delay);
        DEPRESSED(self);
    }

    if (f_reached) {
        while (params->timer == 0 || get_millis() < finished) {
            delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis()) / portTICK_PERIOD_MS;
            xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS | INPUT_EVENT_BITS, pdTRUE, pdFALSE, delay);
            DEPRESSED(self);
        }
    }
    terminate_timer(self);
}

void launch_timer_after_input_task(uint16_t timer, int starter, uint8_t limit420ma, int8_t num420ma, int direction,
        uint8_t *inputbitmap, uint8_t *outputbitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, outputbitmap,  TOT_OUTPUTS);
    memcpy(args->inputbitmap, inputbitmap,  TOT_INPUTS);
    args->queue = queue;
    args->eventgroup = group;
    args->timer = timer;
    args->f_dac = 0;
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->starter = starter;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(outputbitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = -1;
    insert_procq(&process_list, process);
    xTaskCreate(task_timer_after_input, "timer", 4096, (void*)process, ASYNC_TASK_PRIORITY+1, &process->task);
    ESP_LOGI(TAG, "created timer process with ma420 limit %i (%i)", args->limit420ma, limit420ma);
    xSemaphoreGive(async_process_sem);
}